#!/usr/bin/env python
# da_initexit.py

"""
Author : peters 

Revision History:
File created on 13 May 2009.

"""

needed_da_items=[
    'time.start',
    'time.finish',
    'time.nlag',
    'time.cycle',
    'dir.da_run',
    'forecast.model',
    'forecast.nmembers',
    'forecast.model.rc',
    'da.system']


import logging
import os
import sys
import shutil
import datetime
import da.tools.rc as rc



class CycleControl(object):
    """
    This object controls a data assimilation cycle. It is created using information from a da rc-file with settings, and
    also using the command line options passed to the job control script that runs in the foreground, or through the queue.
    This object is passed around throughout the code as it also contains information on directory structures, requested time
    windows, and other resources.
    """

    def __init__(self,opts,args):
        """
        Initialization occurs from passed options and arguments to the job control script.
        """

        self.LoadRc(args['rc'])
        self.ValidateRC()

        # Add some useful variables to the rc-file dictionary

        self.da_settings['jobrcfilename']    = args["jobrcfilename"]
        self.da_settings['log']              = args["logfile"]
        self.da_settings['dir.da_submit']    = os.getcwd()
        self.da_settings['da.crash.recover'] = '-r' in opts
        self.da_settings['verbose']          = '-v' in opts
        self.DaSystem                        = None # to be filled later

    def __str__(self):
        """
        String representation of a CycleControl object
        """

        msg = "==============================================================="    ; print msg
        msg = "DA Cycle rc-file is %s" % self.RcFileName                                ; print msg
        msg = "DA Cycle log file is %s" % self.da_settings['log']                       ; print msg
        msg = "DA Cycle run directory is %s" % self.da_settings['dir.da_run']           ; print msg
        msg = "DA Cycle inverse system is %s" % self.da_settings['da.system']           ; print msg
        msg = "DA Cycle forecast model is %s" % self.da_settings['forecast.model']      ; print msg
        msg = "==============================================================="    ; print msg

        return ""


    def LoadRc(self,RcFileName):
        """ 
        This method loads a DA Cycle rc-file with settings for this simulation 
        """

        self.da_settings    = rc.read(RcFileName)
        self.RcFileName     = RcFileName
        self.DaRcLoaded    = True

        msg                 = 'DA Cycle rc-file (%s) loaded successfully'%self.RcFileName ; logging.info(msg)

        return True


    def ValidateRC(self):
        """ 
        Validate the contents of the rc-file given a dictionary of required keys
        """
        from da.tools.general import ToDatetime

        for k,v in self.da_settings.iteritems():
            if v == 'True' : self.da_settings[k] = True
            if v == 'False': self.da_settings[k] = False
            if 'date' in k : self.da_settings[k] = ToDatetime(v)
            if 'time.start' in k : 
                self.da_settings[k] = ToDatetime(v)
            if 'time.finish' in k : 
                self.da_settings[k] = ToDatetime(v)

        for key in needed_da_items:

            if not self.da_settings.has_key(key):
                status,msg = ( False,'Missing a required value in rc-file : %s' % key)
                logging.error(msg)
                raise IOError,msg

        status,msg = ( True,'DA Cycle settings have been validated succesfully' )  ; logging.debug(msg)

        return None

    def ParseTimes(self):
        """ 
        Parse time related parameters into datetime objects for later use 
        """
        from da.tools.general import AdvanceTime

        startdate = self.da_settings['time.start']
        finaldate = self.da_settings['time.finish']                  

        if finaldate <= startdate:
            msg   = 'The start date (%s) is not greater than the end date (%s), please revise'%(startdate.strftime('%Y%m%d'),finaldate.strftime('%Y%m%d')) 
            logging.error(msg)
            raise ValueError
        #
        cyclelength = self.da_settings['time.cycle']                 # get time step

# Determine end date

        if cyclelength == 'infinite':
            enddate = finaldate
        else:
            enddate = AdvanceTime(startdate,cyclelength)

        #
        if enddate > finaldate:  # do not run beyon finaldate
            enddate = finaldate

        self.da_settings['time.start']   = startdate
        self.da_settings['time.end']     = enddate
        self.da_settings['time.finish']   = finaldate
        self.da_settings['cyclelength'] = cyclelength

        msg = "==============================================================="    ; logging.info(msg)
        msg = "DA Cycle start date is %s"   % startdate.strftime('%Y-%m-%d %H:%M')      ; logging.info(msg)
        msg = "DA Cycle end   date is %s"   % enddate.strftime('%Y-%m-%d %H:%M')        ; logging.info(msg)
        msg = "DA Cycle final date is %s"   % finaldate.strftime('%Y-%m-%d %H:%M')      ; logging.info(msg)
        msg = "DA Cycle cycle length is %s" % cyclelength                               ; logging.info(msg)
        msg = "DA Cycle restart is %s"      % str(self.da_settings['time.restart'])     ; logging.info(msg)
        msg = "==============================================================="    ; logging.info(msg)

        return None

    def RandomSeed(self,action='read'):
        """ Get the randomseed and save it, or read the random seed and set it"""
        import cPickle
        import numpy as np

        filename = os.path.join(self.da_settings['dir.save'],'randomseed.pickle')

        if action == 'write':
            f        = open(filename,'wb')
            seed     = np.random.get_state()
            dummy    = cPickle.dump (seed,f,-1)
            dummy    = f.close()

            msg      = "Saved the random seed generator values to file"

        if action == 'read':
            f        = open(filename,'rb')
            seed     = cPickle.load(f)
            dummy    = np.random.set_state(seed)
            dummy    = f.close()

            msg      = "Retrieved the random seed generator values from file"

        logging.info(msg)

        return None

    def Initialize(self):
        """ Determine how to proceed with this cycle: 
        
            (a) recover from crash : get the save data from the one-ago folder and replace the da_runtime variables with those from the save dir
            (b) restart cycle      : use latest da_runtime variables from the exec dir
            (c) fresh start        : set up the required file structure for this simulation
        """        

#
# case 1: A recover from a previous crash, this is signaled by flag "-r"
#
        if self.da_settings['da.crash.recover']:

            msg   = "Recovering simulation from data in: %s" % self.da_settings['dir.da_run']  ; logging.info(msg)

            dummy = self.SetupFileStructure()

            dummy = self.RecoverRun()

            dummy = self.RandomSeed('read')
#
# case 2: A continuation, this is signaled by rc-item time.restart = True
#
        elif self.da_settings['time.restart']:

            msg   = "Restarting filter from previous step"  ; logging.info(msg)

            dummy = self.SetupFileStructure()

            dummy = self.RandomSeed('read')
#
# case 3: A fresh start, this is signaled by rc-item time.restart = False
#
        elif not self.da_settings['time.restart']:
            msg   = "First time step in filter sequence"   ; logging.info(msg)

            dummy = self.SetupFileStructure()


        self.ParseTimes()
        self.WriteRC(self.da_settings['jobrcfilename'])

        return None


    def SetupFileStructure(self):
        """ 
        Create file structure needed for data assimilation system.
            In principle this looks like:

                ${da_rundir}
                ${da_rundir}/input
                ${da_rundir}/output
                ${da_rundir}/exec
                ${da_rundir}/diagnostics
                ${da_rundir}/analysis
                ${da_rundir}/jobs
                ${da_rundir}/save
                ${da_rundir}/save/one-ago
                ${da_rundir}/save/two-ago

            Note that the exec dir will actually be a simlink to the directory where
            the SIBCASA or model executable lives. This directory is passed through
            the da.rc file. The observation input files will be placed in the exec dir,
            and the resulting simulated values will be retrieved from there as well.
        """
        from da.tools.general import CreateDirs

# Create the run directory for this DA job, including I/O structure

        filtertime  = self.da_settings['time.start'].strftime('%Y%m%d')

        CreateDirs(self.da_settings['dir.da_run'])

        self.da_settings['dir.exec']        = os.path.join(self.da_settings['dir.da_run'],'exec')
        self.da_settings['dir.input']       = os.path.join(self.da_settings['dir.da_run'],'input')
        self.da_settings['dir.output']      = os.path.join(self.da_settings['dir.da_run'],'output','%s'%filtertime)
        self.da_settings['dir.diagnostics'] = os.path.join(self.da_settings['dir.da_run'],'diagnostics')
        self.da_settings['dir.analysis']    = os.path.join(self.da_settings['dir.da_run'],'analysis')
        self.da_settings['dir.jobs']        = os.path.join(self.da_settings['dir.da_run'],'jobs')
        self.da_settings['dir.save']        = os.path.join(self.da_settings['dir.da_run'],'save')

        CreateDirs(os.path.join(self.da_settings['dir.exec']))
        CreateDirs(os.path.join(self.da_settings['dir.input']))
        CreateDirs(os.path.join(self.da_settings['dir.output']))
        CreateDirs(os.path.join(self.da_settings['dir.diagnostics']))
        CreateDirs(os.path.join(self.da_settings['dir.analysis']))
        CreateDirs(os.path.join(self.da_settings['dir.jobs']))
        CreateDirs(os.path.join(self.da_settings['dir.save']))
        CreateDirs(os.path.join(self.da_settings['dir.save'],'one-ago'))
        CreateDirs(os.path.join(self.da_settings['dir.save'],'two-ago'))

        msg = 'Succesfully created the file structure for the assimilation job'  ; logging.info(msg)


    def RecoverRun(self):
        """Prepare a recovery from a crashed run. This consists of: 
        
                - copying all data from the save/one-ago folder,
                - replacing all rc-file items with those from da_runtime.rc 
                - resetting the seed of the random number generator to the value it had before the crash
                
        """

        # Move all data from the save/one-ago directory to the save/current directory

        dummy = self.MoveSaveData(io_option='restore',save_option='full',filter=[])

        # Replace rc-items with those from the crashed run last rc-file

        file_rc_rec = os.path.join(self.da_settings['dir.save'],'da_runtime.rc')
        rc_rec      = rc.read(file_rc_rec)

        for k,v in rc_rec.iteritems():
            self.da_settings[k] = v

        self.ValidateRC()

        msg         = "Replaced rc-items.... "                                        ; logging.debug(msg)
        msg         = "Next cycle start date is %s" % self.da_settings['time.start']  ; logging.debug(msg)

        return None

    def Finalize(self):
        """
        Finalize the da cycle, this means writing the save data and rc-files for the next run
        """

        dummy = self.RandomSeed('write')
        dummy = self.WriteNewRCfile()
        dummy = self.CollectSaveData()
        dummy = self.MoveSaveData(io_option='store',save_option='full')
        dummy = self.SubmitNextCycle()

    def CollectSaveData(self):
        """ Collect files needed for the restart of this cycle in case of a crash, or for the continuation of the next cycle. 
            All files needded are written to the save/ directory.
            Currently, the following files are included

            -) The randomseed.pickle file, also already in the save directory
            -) The da_runtime.rc file
            -) The savestate.yyyymmdd.nc file
            -) ...

            Note that we assume that the transport model restart files are directly written to the save/ directory and do not need to be collected

        """

        filtertime  = self.da_settings['time.start'].strftime('%Y%m%d')

        file1       = os.path.join(self.da_settings['dir.exec'],'da_runtime.rc')
        file2       = os.path.join(self.da_settings['dir.output'],'savestate.nc')

        FilesToSave = [file1,file2]

        targetdir   = os.path.join(self.da_settings['dir.save'])

        msg         = "Collecting the required restart/save data"                                 ; logging.info(msg)
        msg         = "           to   directory: %s " % targetdir                                ; logging.debug(msg)

        for file in FilesToSave:

            msg         = "           [copy] .... %s " % file                                     ; logging.debug(msg)
            dummy       = shutil.copy(file,file.replace(os.path.split(file)[0],targetdir) )


    def MoveSaveData(self, io_option='restore', save_option='partial',filter=[]):
        """ Store or restore model state to/from a temporary directory. Two save_option options are available:

            (1) save_option = 'partial'  : Save the background concentration fields only, to use for a later advance of the background
            (2) save_option = 'full'     : Save all restart data, to use for a next cycle

            While also two IO options are available:

            (1) io_option = restore    : Get data from a directory
            (2) io_option = store      : Save data to a directory

            In case of a 'store' command the target folder is re-created so that the contents are empty to begin with.

            There is currently room for the specific sub-models to also add save data to the mix. This is done through a 
            forecast.model dependend import statement, followed by a call to function:

            model.MoveSaveData(**args)

        """
        from da.tools.general import CreateDirs

        if io_option not in ['store','restore']:
            raise ValueError,'Invalid option specified for io_option (%s)' % io_option
        if save_option not in ['partial','full']:
            raise ValueError,'Invalid option specified for save_option (%s)' % save_option


        savedir     = os.path.join(self.da_settings['dir.save'])
        oneagodir   = os.path.join(self.da_settings['dir.save'],'one-ago')
        tempdir     = os.path.join(self.da_settings['dir.save'],'tmp')


        if save_option == 'partial':   # save background data to tmp dir

            if io_option == 'store':

                targetdir = tempdir
                sourcedir = savedir

            elif io_option == 'restore':

                sourcedir = tempdir
                targetdir = savedir

        elif save_option == 'full':   # move all save data to one-ago dir

            if io_option == 'store':

                targetdir = oneagodir
                sourcedir = savedir

            elif io_option == 'restore':

                sourcedir = oneagodir
                targetdir = savedir

# If "store" is requested, recreate target dir, cleaning the contents 

        if io_option == 'store':
            CreateDirs(os.path.join(targetdir),forceclean=True)

        msg         = "Performing a %s %s of data" % (save_option,io_option)                      ; logging.debug(msg)
        msg         = "           from directory: %s " % sourcedir                                ; logging.debug(msg)
        msg         = "           to   directory: %s " % targetdir                                ; logging.debug(msg)
        msg         = "           with filter   : %s " % filter                                   ; logging.debug(msg)


        for file in os.listdir(sourcedir):

            file = os.path.join(sourcedir,file)

            if os.path.isdir(file): # skip dirs
                skip = True
            elif filter == []:      # copy all
                skip= False        
            else:                   # check filter
                skip = True         # default skip
                for f in filter:
                    if f in file: 
                        skip = False # unless in filter
                        break
                
            if skip: 
                msg         = "           [skip] .... %s " % file                                     ; logging.debug(msg)
                continue    


            if io_option == 'store' and save_option == 'full':
                #msg         = "           [move] .... %s " % file                                                ; logging.debug(msg)
                #dummy       = shutil.move(file,file.replace(sourcedir,targetdir) )
                msg         = "           [copy] .... %s " % file                                                ; logging.debug(msg)
                dummy       = shutil.copy(file,file.replace(sourcedir,targetdir) )

            else:        
                msg         = "           [copy] .... %s " % file                                                ; logging.debug(msg)
                dummy       = shutil.copy(file,file.replace(sourcedir,targetdir) )

#
    def WriteNewRCfile(self):
        """ Write the rc-file for the next DA cycle. Note that the start time for the next cycle is the end time of this one, while 
            the end time for the next cycle is the current end time + one cycle length. The resulting rc-file is written to:
            
            (1) to the dir.exec so that it can be used when resubmitting the next cycle
            
        """
        from da.tools.general import AdvanceTime

        # These first two lines advance the filter time for the next cycle

        self.da_settings['time.start']        = self.da_settings['time.end']
        self.da_settings['time.end']          = AdvanceTime(self.da_settings['time.end'],self.da_settings['cyclelength'])

        # The rest is info needed for a system restart

        self.da_settings['time.restart']     = True
        #self.da_settings['time.start']       = self.da_settings['time.start'].strftime('%Y-%m-%d %H:%M:%S')
        #self.da_settings['time.finish']      = self.da_settings['time.finish'].strftime('%Y-%m-%d %H:%M:%S')

        fname                                = os.path.join(self.da_settings['dir.exec'],'da_runtime.rc')
        self.da_settings['da.restart.fname'] = fname
        dummy                                = rc.write(fname,self.da_settings)

        msg = 'Wrote new da_runtime.rc file to working directory' ; logging.debug(msg)


    def WriteRC(self,filename):
        """ Write RC file after each process to reflect updated info """

        rcname = filename
        fname  = './'+rcname
        dummy  = rc.write(fname,self.da_settings)
        msg    = 'Wrote expanded rc-file (%s) to current directory (%s)'%(rcname,self.da_settings['dir.exec']) ; logging.debug(msg)
        return None

    def SubmitNextCycle(self):
        """ 
        Submit next job in cycle, if needed. The end date is held against the final date. A subprocess is called that 
        submits the next job, without waiting for its result so this cycle can finish and exit. 
        """
        import subprocess
        import os


        if self.da_settings['time.start'] < self.da_settings['time.finish']:
            #code  = subprocess.call(['qsub',sys.argv[0],'rc=%s' % rc_da_shell['da.restart.fname']] )
            code  = subprocess.Popen([os.getcwd()+'/das.py','rc=%s' % self.da_settings['da.restart.fname'],"process=all"] )
            code=0
            if code == 0:
                logging.info('Submitted next cycle succesfully ')
            else:
                logging.error('Error in submitting next cycle, return code: %s ' % code)
                raise OSError
        else:
            logging.info('Final date reached, no new cycle started')

        return None

def StartLogger(logfile='jb.%s.log'%os.getpid()):
    """ start the logging of messages to screen and to file"""

# start the logging basic configuration by setting up a log file

    logging.basicConfig(level    = logging.DEBUG,
                        format   = ' [%(levelname)-7s] (%(asctime)s) py-%(module)-20s : %(message)s',
                        datefmt  = '%Y-%m-%d %H:%M:%S',
                        filename = logfile,
                        filemode = 'a')

# define a Handler which writes INFO messages or higher to the sys.stdout

    console = logging.StreamHandler(sys.stdout)
    dummy   = console.setLevel(logging.INFO)

# set a format which is simpler for console use

    formatter = logging.Formatter('[%(levelname)-7s] (%(asctime)s) py-%(module)-20s: %(message)s')

# tell the handler to use this format

    dummy = console.setFormatter(formatter)

# add the handler to the root logger

    dummy = logging.getLogger('').addHandler(console)


def ParseOptions():
    """ Function parses options from the command line and returns the arguments as a dictionary"""
    import getopt
    import sys

# Parse keywords, the only option accepted so far is the "-h" flag for help

    opts=[]
    args=[]
    try:                                
        opts, args = getopt.gnu_getopt(sys.argv[1:], "-hrv")
    except getopt.GetoptError, msg:           
        logging.error('%s'%msg)
        sys.exit(2)      

    for options in opts:
        options=options[0].lower()
        if options == '-h':
            print ""
            print helptext
            sys.exit(2)      
        if options == '-r':
            logging.info('-r flag specified on command line: recovering from crash')
        if options == '-v':
            logging.info('-v flag specified on command line: extra verbose output')
            dummy  = logging.root.handlers[-1].setLevel(logging.DEBUG)

# Parse arguments and return as dictionary

    arguments={}
    for item in args:
        #item=item.lower()

# Catch arguments that are passed not in "key=value" format

        if '=' in item:
            key, arg = item.split('=')
        else:
            logging.error('%s'%'Argument passed without description (%s)' % item)
            raise getopt.GetoptError,arg

        arguments[key]=arg

    if opts: opts=opts[0]

    return opts, arguments

def ValidateOptsArgs(opts,args):
    """ Validate the options and arguments passed from the command line before starting the cycle """

    if not args.has_key("rc"):
        msg = "There is no rc-file specified on the command line. Please use rc=yourfile.rc"   ; logging.error(msg)
        raise IOError,msg
    elif not os.path.exists(args['rc']):
        msg = "The specified rc-file (%s) does not exist " % args['rc'] ;  logging.error(msg)
        raise IOError,msg

    args['jobrcfilename'] = "jb.%s.rc"%(os.getpid(),)
    args['logfile']       = "jb.%s.log"%(os.getpid(),)


    return opts,args

def CleanUpCycle(DaCycle):
    """
    Move files, clean up rundir after a cycle has finished
    """

    # move log file to rundir/jobs
    jobdir      = os.path.join(DaCycle.da_settings['dir.da_run'],"jobs")
    logfile     = DaCycle.da_settings['log']
    joblogfile  = os.path.join(jobdir,logfile) 
    dummy       = shutil.move(logfile,joblogfile)
    msg         = "....Moved %s to %s"%(logfile,joblogfile)   ; logging.debug(msg)

    # move rc file to rundir/jobs
    jobrcfile   = os.path.join(jobdir,DaCycle.da_settings["jobrcfilename"] )
    dummy       = shutil.move(DaCycle.da_settings["jobrcfilename"],jobrcfile )
    msg         = "....Moved %s to %s"%(DaCycle.da_settings['jobrcfilename'],jobrcfile)   ; logging.debug(msg)

    # WP TOO SPECIFIC NOW, MAKE A ROUTINE THAT CLEANS UP ANY MODEL OUTPUT FILE, NOT JUST TM5!!!

    # cat model output and rc-file to the job file output
    tm5jobfile  = os.path.join(jobdir,"tm5.%s"%(DaCycle.da_settings['log']) )
    if  os.path.exists(tm5jobfile):
        msg     = "....Concatenating %s to %s"%(tm5jobfile,joblogfile)   ; logging.debug(msg)
        f       = open(joblogfile,'a')
        dummy   = f.write(open(tm5jobfile,'r').read())
        dummy   = f.close()
    if  os.path.exists(jobrcfile):
        msg     = "....Concatenating %s to %s"%(jobrcfile,joblogfile)   ; logging.debug(msg)
        f       = open(joblogfile,'a')
        dummy   = f.write(open(jobrcfile,'r').read())
        dummy   = f.close()
    msg = "The complete log file is now at: %s"%(joblogfile)   ; logging.info(msg)



if __name__ == "__main__":

    sys.path.append('../../')

    from da.tools.general import StartLogger 

    opts = ['-v']
    args = {'rc':'../../da.rc','logfile':'da_initexit.log','jobrcfilename':'test.rc'}

    StartLogger()
    DaCycle = CycleControl(opts,args)
    DaCycle.Initialize()
    DaCycle.Finalize()
    print DaCycle

    #opts = ['-v','-r']
    #DaCycle = CycleControl(opts,args)
    #DaCycle.Initialize()
    #DaCycle.Finalize()
    #print DaCycle

